
/* ACUA RADIUS Accountancy Daemon */
/* Version 0.01 */
/* Modified Last: 20-11-2000 */

#include <stdio.h>              // file functions
#include <stdlib.h>             // malloc/free/strtoul
#include <string.h>             // strncpy
#include <ctype.h>              // isalpha
#include <limits.h>             // declarations: ULONG_MAX
#include <errno.h>              // declarations: errno, ERANGE
#include <syslog.h>             // syslog
#include <unistd.h>             // fcntl.h
#include <fcntl.h>              // fcntl
#include <sys/file.h>           // open, close and fdopen
#include <sys/stat.h>           // fstat and stat structure.
#include <sys/types.h>
#include <time.h>

#include "common.h"             // ACUA
#include "radacct.h"            // RADIUS Accountancy.



void strip_quote(char *s) {
    char  *tmp;
    unsigned int   i = 0, count = 0;
    // how many occourances of '"' are there?
    for (i=0;i<strlen(s);i++)
	if (s[i] == '"') count++;
    // allocate enough memory for our output string
    tmp = (char *) malloc(strlen(s)-count+1);
    count = 0;
    // copy the string, but skip all of the '"' charactors.
    for (i=0;i<strlen(s);i++)
	if (s[i] != '"')
	    tmp[count++] = s[i];
    // terminate string, copy the string, and free the malloc()'d space
    tmp[count++] = '\0';
    strncpy(s, tmp, count);
    free(tmp);
}

int check_strtoul(unsigned long int uli) {
    /* Purpose:
     *   Check the error condition from the last call to strtoul()
     * Return Values:
     *   0 = Success.
     *   1 = Failed.
     */
    if ((uli == ULONG_MAX) && (errno == ERANGE))
	return(1);
    else
	return(0);
}

int detailFileOpen(char *path) {
    /* Purpose:
     *   Open the detail file in read-only mode.
     * Return Values:
     *   0 = Success.
     *   1 = Failed.
     */

    if (!(acctFilePtr = fopen(path, "r")))
	return(1);

    /* Open was successful. */
    return(0);
}

int detailFileClose(void) {
    /* Purpose:
     *   Close the detail file.
     * Return Values:
     *   0 = Success.
     *   1 = Failed.
     */
    if (fclose(acctFilePtr))
	return(1);
    /* Close was successful. */
    return(0);
}

int get_next_entry(radacct *acct, fpos_t *fpos) {
    /* Purpose:
     *   Retrieve the next record from the detail file.
     * Return Values:
     *   0 = Success.
     *   1 = End of file reached.
     *   2 = User-Name value too long.
     *   3 = Corrupt record.
     */
    char          tmp[80],
    attr[80],
    crap[80],
    value[80];
    char          *res;

    attr[0] = '\0';
    // clear acct structure
    acct->session_id[0] = '\0';
    acct->user_name[0] = '\0';
    acct->port_id = 0;
    acct->in_octets = 0;
    acct->out_octets = 0;

    // seek to fpos
    fsetpos(acctFilePtr, fpos);

    if (feof(acctFilePtr)) return(1);  // end of file

    for (;;) {
	if (feof(acctFilePtr)) return(1);  // end of file

	// FIXME: REPLACE THIS LINE WITH A 'getline' CALL!!
	// remember to check for EOF though - this is _vital_ because this is
	// where we just give up and dont bother trying to read this record any
	// further until a minute later or so.
	res = fgets(tmp, sizeof(tmp), acctFilePtr);
	if (!res) return(1);   // probably EOF, not 100% reliable though.

	// test for the end of the record.
	if (tmp[0] == 10) {                 // end of record
	    fgetpos(acctFilePtr, fpos);     // record new file position
	    return(0);
	}

	if (strlen(tmp) <= 1) continue;    // dont process.

	/* FIXME: dont process, this is most likely the date line.
	 * We could possibly throw it in our struct, just for good measure.
	 */
	if (isalpha(tmp[0])) continue;

	/* Extract the info from the line of text */
	sscanf(tmp, "%s %s %s", attr, crap, value);

	/* condition the text */
	strip_quote(value);

	if (!strcmp(attr, "Acct-Session-Id")) {
	    if (sizeof(acct->session_id) >= strlen(value)+1)
		strncpy(acct->session_id, value, strlen(value)+1);
	    else {
		syslog(LOG_ERR, "Acct-Session-Id value is too long!");
		acct->session_id[0] = '\0';
		return(2);
	    }
	}
	else if (!strcmp(attr, "User-Name"))
	    if (sizeof(acct->user_name) >= strlen(value)+1)
		strncpy(acct->user_name, value, strlen(value)+1);
	    else {
		syslog(LOG_ERR, "User-Name value is too long!");
		acct->user_name[0] = '\0';
		return(2);
	    }
	else if (!strcmp(attr, "Acct-Status-Type")) {
	    if (!strcmp(value, "Stop"))
		acct->rectype = 0;
	    else if (!strcmp(value, "Start"))
		acct->rectype = 1;
	    else return(3);
	}
	else if (!strcmp(attr, "NAS-Port-Id")) {
	    acct->port_id = strtoul(value, (char **)NULL, 10);
	    if (check_strtoul(acct->port_id)) return(3);
	}
	else if (!strcmp(attr, "Acct-Input-Octets")) {
	    acct->in_octets = strtoul(value, (char **)NULL, 10);
	    if (check_strtoul(acct->in_octets)) return(3);
	}
	else if (!strcmp(attr, "Acct-Output-Octets")) {
	    acct->out_octets = strtoul(value, (char **)NULL, 10);
	    if (check_strtoul(acct->out_octets)) return(3);
	}
    }
    return(0);
}

void usage(char **argv)
{
    printf("usage:\n");
    printf("  %s [-d] [-e] <path to detail file>\n\n", argv[0]);
    printf("  -d = Do not go into daemon mode.\n");
    printf("  -e = Do not save changes to the user database.\n");
    exit(1);
}


int main(int argc, char **argv)
{
    /* Purpose:
     *   Test all routines in this (to be) API.
     * Return Values:
     *   0 = All OK.
     *   1 = Error.
     */
    int         	res, runAsDaemon, effect;
    char        	*detail_file;
    struct	stat  	detail_stat_orig, detail_stat_cur;
    time_t 		clkinfo;
    struct      tm      *tmP;

    radacct     	acct;
    fpos_t      	fpos;

    UserRec     	ur;
    uid_t       	uid;

    /* Tempory syslog stuff for debugging purposes */
    openlog("acua_radacctd", LOG_PID, LOG_DAEMON);
    effect = 1;
    runAsDaemon = 1;

    /* See if we have the minimum amount of command line options */
    if (argc < 2)
        usage(argv);

    while ((res = getopt(argc, argv, "de")) != EOF) {
	switch (res) {
	case 'd':
	    runAsDaemon = 0;
	    break;
	case 'e':
	    effect = 0;
	    break;
	}
    }

    if (optind == argc)
	usage(argv);

    /* become a daemon if desired */
    if (runAsDaemon) daemonInit();

    /* Allocate the detail_file variable */
    detail_file = (char *) malloc(strlen(argv[optind]) + 1);
    strncpy(detail_file, argv[optind], strlen(argv[optind]));

    /* Open the detail file */
    if (detailFileOpen(argv[optind])) {
	printf("Error opening detail file: %s\n", detail_file);
	exit(1);
    }

    /* record some info about the file, so that we can tell if it's been
     * rotated or not.
     */
    if (stat(detail_file, &detail_stat_orig))
    {
	syslog(LOG_ERR, "Could not stat file: %s", detail_file);
	exit(1);
    }

    // seek to end of file
    fseek(acctFilePtr, 0, SEEK_END);
    fgetpos(acctFilePtr, &fpos);

    for (;;) {

	/* this number is obscure on purpose to help prevent radupdated
	 * and radacctd from clashing when accessing the user database.
         */
	sleep(33);

        /* Check to see if the file size is now less than it was before,
	 * if so, we need to close and re-open the file because chances are
	 * the log has been rotated.
	 */
	if (stat(detail_file, &detail_stat_cur))
	{
	    syslog(LOG_ERR, "Could not stat file: %s", detail_file);
	    exit(1);
	}

	if (detail_stat_cur.st_ino != detail_stat_orig.st_ino)
	{
	    /* Detail file has been rotated */
	    detailFileClose();
            /* Open the detail file */
	    if (detailFileOpen(argv[optind])) {
		syslog(LOG_ERR, "Error opening detail file: %s\n", detail_file);
		exit(1);
	    }
	    if (stat(detail_file, &detail_stat_orig))
	    {
		syslog(LOG_ERR, "Could not stat file: %s", detail_file);
		exit(1);
	    }
            /* Reset the file position indicator */
	    fgetpos(acctFilePtr, &fpos);
	}


	for (;;)
	{
	    res = get_next_entry(&acct, &fpos);

	    /* No more records to read. */
	    if (res == 1)
		break;

            /* Make sure we only process Stop records. */
	    if (acct.rectype)
                break;

	    if (res == 0)
	    {
		//printf("Entry read ok.\n");
		//printf("Session ID: %s\n", acct.session_id);
		//printf("Login Name: %s\n", acct.user_name);
		//printf("In  Octets: %lu\n", acct.in_octets);
		//printf("Out Octets: %lu\n", acct.out_octets);
		//printf("Rec   Type: %d\n", acct.rectype);
		//printf("---\n");

		/* FIXME: Ok, I recon this is inefficient to be opening and
		 * closing the user database file all the time, but it's probably
		 * safer, atleast for now, as this is pretty new code, and the
		 * get_next_entry() call might have a bug that could cause it to
		 * effectivly hang, thus a lock could be present on the particular user
		 * recond, and would crash acua_radupdated until it's released.
		 */

		/* Get the UID from the username. */
		uid = UIDfromLogin(acct.user_name);

		/* Check if we found a UID to corrospond to the username, if so
		 * we can then go ahead and update the user record.
		 */
		if (uid == (uid_t) -1)
		    break;

		/* FIXME: error checking should be implemented in libacua.cc. */
		userFileOpen();

		/* Find the ACUA user record.  If it does not exist then close the
		 * user database and look for a new RADIUS record.
		 */
		if (userFileSearch(&ur, uid))
		{
		    userFileClose();
		    break;
		}

		/* Everything looks sweet.  Update the user record. */
		ur.bTx += acct.out_octets;
		ur.bRx += acct.in_octets;
		ur.bStx = acct.out_octets;
		ur.bSrx = acct.in_octets;

		/* record monthly transfer stats */
		if (time(&clkinfo) != ((time_t)-1) &&
		    (tmP = localtime(&clkinfo)))
		{
		    ur.monthbTx[tmP->tm_mon] += acct.out_octets;
		    ur.monthbRx[tmP->tm_mon] += acct.in_octets;
		}

		/* Tempory debugging information */
		syslog(LOG_INFO, "[%s] Downloaded: %li, Uploaded %li.", acct.user_name, acct.out_octets, acct.in_octets);

		/* Write the changes. */
		if (effect)
		    userFileEdit(&ur);

		/* Close the user file.
		 * FIXME: error checking should be implemented.
		 */
		userFileClose();
	    }

	    // FIXME: Eliminate this possible error.
	    if (res == 2)
	    {
		printf("User-Name or Acct-Session-Id value too large - Aborting.\n");
		break;
	    }

	    if (res == 3)
	    {
		printf("Corrupt record.\n");
		/* FIXME: wack a thing in here to seek to the end of the
		 * corrupt record and update the file position indicator
		 */
		break;
	    }
	}
    }

    detailFileClose();
    return(0);
}

